package cz.fidentis.analyst.procrustes;

import cz.fidentis.analyst.face.HumanFace;
import cz.fidentis.analyst.feature.FeaturePoint;
import cz.fidentis.analyst.feature.api.IPosition;
import org.ejml.simple.SimpleMatrix;
import org.ejml.simple.SimpleSVD;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.DataFormatException;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Procrustes analysis computing transformation of a face.
 * <p>
 * Superimposes two faces over each other so their all of the feature points are
 * as close to their corespondent type from the other face as they can be
 * without changing the general shape itself.
 * <p>
 * Transformations that class computes are adjustment by vector, scaling and
 * rotation of feature points. By computing these values they can later be
 * applied on the face (vertices of face).
 * <p>
 * For more information take a look at:
 * https://en.wikipedia.org/wiki/Procrustes_analysis
 *
 * @author Jakub Kolman
 */
public class ProcrustesAnalysis {

    private ProcrustesAnalysisFaceModel faceModel1;
    private ProcrustesAnalysisFaceModel faceModel2;

    protected final Point3d modelCentroid1;
    protected final Point3d modelCentroid2;

    private boolean scale = false;

    /**
     * Constructor
     *
     * @param humanFace1
     * @param humanFace2
     * @throws DataFormatException
     */
    public ProcrustesAnalysis(
            HumanFace humanFace1,
            HumanFace humanFace2) {

        ProcrustesAnalysisFaceModel model1;
        ProcrustesAnalysisFaceModel model2;

        // check if feature points lists of both faces have the same size and same types
        // if not, than we will be working with subset or cancel the analysis
        if (humanFace1.getFeaturePoints().size() != humanFace2.getFeaturePoints().size()
                || !ProcrustesUtils.checkFeaturePointsType(
                        ProcrustesUtils.sortListByFeaturePointType(humanFace1.getFeaturePoints()),
                        ProcrustesUtils.sortListByFeaturePointType(humanFace1.getFeaturePoints()))) {
            ArrayList<Integer> viablePoints = getSubsetOfFeaturePoints(humanFace1.getFeaturePoints(), humanFace2.getFeaturePoints());
            model1 = new ProcrustesAnalysisFaceModel(humanFace1, viablePoints);
            model2 = new ProcrustesAnalysisFaceModel(humanFace2, viablePoints);
        } else {
            model1 = new ProcrustesAnalysisFaceModel(humanFace1);
            model2 = new ProcrustesAnalysisFaceModel(humanFace2);
        }
        this.modelCentroid1 = findCentroidOfFeaturePoints(model1.getFeaturePointValues());
        this.modelCentroid2 = findCentroidOfFeaturePoints(model2.getFeaturePointValues());
        this.faceModel1 = model1;
        this.faceModel2 = model2;
    }

    /**
     * Constructor with variable for scaling option. Default scale option is set
     * to false.
     *
     * @param humanFace1
     * @param humanFace2
     * @param scale
     * @throws DataFormatException
     */
    public ProcrustesAnalysis(
            HumanFace humanFace1,
            HumanFace humanFace2,
            boolean scale) throws DataFormatException {
        this(humanFace1, humanFace2);
        this.scale = scale;
    }

    /**
     * Method called for analysis after creating initial data in constructor.
     * This method causes superimposition and rotation of the faces.
     *
     * Using generalized orthogonal procrustes analysis need to be first moved
     * over each other, then scale and then rotate.
     *
     * @throws DataFormatException if faces have less than 3 feature points in
     * which case analysis doesn't make sense
     */
    public ProcrustesTransformation analyze() throws DataFormatException {
        ProcrustesTransformation transformation = new ProcrustesTransformation();
        if (this.faceModel1.getFeaturePointsMap().size() >= 3) {
            // adjustment is used for superimposition and return to the original place
            transformation.setCentroidAdjustment(new Vector3d(-this.modelCentroid2.x, -this.modelCentroid2.y, -this.modelCentroid2.z));
            transformation.setSuperImpositionAdjustment(new Vector3d(this.modelCentroid1.x, this.modelCentroid1.y, this.modelCentroid1.z));

            // sets both faces to the origin of plains
            superImpose();

            // calculation of scaling vector
            if (scale) {
                double scaleFactor = this.calculateScalingValue();
                transformation.setScale(scaleFactor);
                calculateScaledList(faceModel2.getFeaturePointValues(), scaleFactor);
            }
            // calculation of rotation matrix
            transformation.setRotationMatrix(this.rotate());

            // move face vertices back, so the original place, 
            // but set feture points map to already superimposed position
            moveFaceModel(faceModel1, new Vector3d(this.modelCentroid1.x, this.modelCentroid1.y, this.modelCentroid1.z));
            moveFaceModel(faceModel2, new Vector3d(this.modelCentroid1.x, this.modelCentroid1.y, this.modelCentroid1.z));

            transformation.setFeaturePointsMap(faceModel2.getFeaturePointsMap());

        } else {
            throw new DataFormatException("Faces have less than 3 feature points.");
        }
        return transformation;
    }

    /**
     * Moves all vertices and feature points of a face by given vector value
     *
     * @param faceModel face that is moved
     * @param vector vector its values are used to move a face
     */
    private void moveFaceModel(ProcrustesAnalysisFaceModel faceModel, Vector3d vector) {
        faceModel.getFeaturePointsMap().values().forEach(fp -> {
            ProcrustesUtils.movePoint(fp, vector);
        });
    }

    /**
     * Moves face model by given vector values. Different vector value is used
     * for vertices and feature points. This method is used to fit feature
     * points to already superimposed position over given face1, but move
     * vertices of face back to original position of face so all the vertex
     * movement is done by procrustes visitor.
     *
     * @param faceModel on which adjustment will be applied
     * @param vertexVector vertex adjustment
     * @param featurePointVector feature points adjustment
     */
    /**
     * By rotation of matrices solves orthogonal procrustes problem.
     */
    private SimpleMatrix rotate() {
        SimpleMatrix primaryMatrix = createMatrixFromList(this.faceModel2.getFeaturePointValues());
        SimpleMatrix transposedMatrix = createMatrixFromList(
                this.faceModel1.getFeaturePointValues()).transpose();

        SimpleMatrix svdMatrix = transposedMatrix.mult(primaryMatrix);
        SimpleSVD<SimpleMatrix> singularValueDecomposition = svdMatrix.svd();
        SimpleMatrix transposedU = singularValueDecomposition.getU().transpose();
        SimpleMatrix rotationMatrix = singularValueDecomposition.getV().mult(transposedU);
        primaryMatrix = primaryMatrix.mult(rotationMatrix);

        this.faceModel2.setFeaturePointsMap(
                createFeaturePointMapFromMatrix(
                        primaryMatrix, this.faceModel2));

        return rotationMatrix;
    }

    /**
     * Calculate scaling ratio of how much the appropriate object corresponding
     * to the second feature point list has to be scale up or shrunk.
     * <p>
     * If returned ratioValue is greater 1 then it means that the second object
     * should be scaled up ratioValue times. If returned ratioValue is smaller 1
     * than the second object should be shrunk.
     *
     * @return ratioValue
     */
    private double calculateScalingValue() {
        List<FeaturePoint> featurePointList1 = this.faceModel1.getFeaturePointValues();
        List<FeaturePoint> featurePointList2 = this.faceModel2.getFeaturePointValues();

        double[] distancesOfList1 = new double[featurePointList1.size()];
        double[] distancesOfList2 = new double[featurePointList2.size()];

        Point3d featurePointCentroid1 = findCentroidOfFeaturePoints(
                featurePointList1);
        Point3d featurePointCentroid2 = findCentroidOfFeaturePoints(
                featurePointList2);

        for (int i = 0; i < featurePointList1.size(); i++) {
            distancesOfList1[i] = calculateDistanceFromPoint(featurePointList1.get(i), featurePointCentroid1);
            distancesOfList2[i] = calculateDistanceFromPoint(featurePointList2.get(i), featurePointCentroid2);
        }

        double[] ratioArray = new double[distancesOfList1.length];
        double ratioValue = 0;

        for (int i = 0; i < distancesOfList1.length; i++) {
            ratioArray[i] += distancesOfList1[i] / distancesOfList2[i];
        }
        for (int i = 0; i < ratioArray.length; i++) {
            ratioValue += ratioArray[i];
        }
        return ratioValue / distancesOfList1.length;
    }

    /**
     * Scales each given point from list by multiplying its position coordinates
     * with scaleFactor.
     *
     * @param list
     * @param scaleFactor
     * @param <T> either MeshPoint or FeaturePoint type.
     */
    private <T extends IPosition> void calculateScaledList(List<T> list, double scaleFactor) {
        List<T> scaledList = new ArrayList<>();
        for (T point : list) {
            scaledList.add(ProcrustesUtils.scalePointDistance(point, scaleFactor));
        }
    }

    /**
     * Finds centroid from given feature point List
     *
     * @param featurePointList
     * @return centroid of feature points (Vector3F)
     */
    private Point3d findCentroidOfFeaturePoints(List<FeaturePoint> featurePointList) {
        float x = 0;
        float y = 0;
        float z = 0;
        for (FeaturePoint fp : featurePointList) {
            x += fp.getX();
            y += fp.getY();
            z += fp.getZ();
        }
        return new Point3d(x / featurePointList.size(), y / featurePointList.size(), z / featurePointList.size());
    }

    /**
     * Creates feature point map HashMap with key FeaturePoint.type and value
     * FeaturePoint back from matrix.
     *
     * @param matrix
     * @param model
     * @return
     */
    private HashMap<Integer, FeaturePoint> createFeaturePointMapFromMatrix(
            SimpleMatrix matrix, ProcrustesAnalysisFaceModel model) {
        HashMap<Integer, FeaturePoint> map = new HashMap<>();
        for (int i = 0; i < matrix.numRows(); i++) {
            FeaturePoint featurePoint = new FeaturePoint(
                    matrix.get(i, 0),
                    matrix.get(i, 1),
                    matrix.get(i, 2),
                    model.getFeaturePointsMap().get(
                            model.getFeaturePointTypeCorrespondence().get(i)
                    ).getFeaturePointType()
            );
            map.put(model.getFeaturePointTypeCorrespondence().get(i), featurePoint);
        }
        return map;
    }

    /**
     * Creates matrix from given feature point list
     *
     * @param list
     * @return matrix
     */
    private <T extends IPosition> SimpleMatrix createMatrixFromList(List<T> list) {
        SimpleMatrix matrix = new SimpleMatrix(list.size(), 3);
        for (int i = 0; i < list.size(); i++) {
            matrix.set(i, 0, list.get(i).getPosition().x);
            matrix.set(i, 1, list.get(i).getPosition().y);
            matrix.set(i, 2, list.get(i).getPosition().z);
        }
        return matrix;
    }

    /**
     * Calculates distance of one feature point from another
     *
     * @param fp1
     * @param point
     * @return distance
     */
    private double calculateDistanceFromPoint(FeaturePoint fp1, Point3d point) {
        return Math.sqrt(
                (Math.pow(fp1.getX() - point.x, 2))
                + (Math.pow(fp1.getY() - point.y, 2))
                + (Math.pow(fp1.getZ() - point.z, 2)));
    }

    /**
     * moves faces so the feature points centroids of both faces will be at the
     * origin (0,0,0)
     */
    protected void superImpose() {
        moveFaceModel(this.faceModel2, new Vector3d(-this.modelCentroid2.x, -this.modelCentroid2.y, -this.modelCentroid2.z));
        moveFaceModel(this.faceModel1, new Vector3d(-this.modelCentroid1.x, -this.modelCentroid1.y, -this.modelCentroid1.z));
    }

    private HashMap<Integer, FeaturePoint> moveFeaturePoints(ProcrustesAnalysisFaceModel faceModel2, Vector3d pointDistance) {
        HashMap<Integer, FeaturePoint> map = new HashMap<>();
        for (Map.Entry<Integer, FeaturePoint> entry : faceModel2.getFeaturePointsMap().entrySet()) {
            entry.getValue().getPosition().x += pointDistance.x;
            entry.getValue().getPosition().y += pointDistance.y;
            entry.getValue().getPosition().z += pointDistance.z;
            map.put(entry.getKey(), entry.getValue());
        }
        return map;
    }

    /**
     * Finds corresponding subset of feature points from both faces so the
     * procrustes analysis can be applied on the subset.
     *
     * @param featurePoints
     * @param featurePoints0
     *
     * @return List of feature point types that are in both sets of feature
     * points
     */
    private ArrayList<Integer> getSubsetOfFeaturePoints(List<FeaturePoint> featurePoints1, List<FeaturePoint> featurePoints2) {
        HashMap<Integer, FeaturePoint> fpMap = new HashMap<>();
        for (FeaturePoint fp : featurePoints1) {
            fpMap.put(fp.getFeaturePointType().getType(), fp);
        }

        ArrayList<Integer> vaiablePoints = new ArrayList<>();
        featurePoints2.forEach(fp -> {
            if (fpMap.get(fp.getFeaturePointType().getType()) != null) {
                vaiablePoints.add(Integer.valueOf(fp.getFeaturePointType().getType()));
            }
        });
        return vaiablePoints;
    }

}
