package cz.fidentis.analyst.procrustes;

import cz.fidentis.analyst.face.HumanFace;
import cz.fidentis.analyst.feature.FeaturePoint;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * Holds important face attributes that are required for procrustes analysis.
 *
 * @author Jakub Kolman
 */
public class ProcrustesAnalysisFaceModel {

//    private final HumanFace humanFace;
    private HashMap<Integer, FeaturePoint> featurePointsMap;        // sorted by feature point type
    private final HashMap<Integer, Integer> featurePointTypeCorrespondence;

    /**
     * constructor creating inner attributes used in procrustes analysis from
     * face
     *
     * @param face
     */
    public ProcrustesAnalysisFaceModel(HumanFace face) {
//        this.humanFace = face;
        // getFeaturePoints() returns unmodifiable List. To sort it we need to make copy first
        List<FeaturePoint> modifiableFeaturePointList = new ArrayList<>(face.getFeaturePoints());

        this.featurePointsMap = createFeaturePointMap(
                ProcrustesUtils.sortListByFeaturePointType(modifiableFeaturePointList));
        this.featurePointTypeCorrespondence = createFeaturePointTypeCorrespondence(face.getFeaturePoints());
    }

    /**
     * Constructor for face model if there is a need for a subset selection of
     * feature points before applying procrustes analysis.
     *
     * @param face
     * @param viablePoints
     */
    public ProcrustesAnalysisFaceModel(HumanFace face, List<Integer> viablePoints) {
//        this.humanFace = face;
        // getFeaturePoints() returns unmodifiable List. To sort it we need to make copy first
        List<FeaturePoint> modifiableFeaturePointList = getSelectionOfFeaturePoints(face.getFeaturePoints(), viablePoints);

        this.featurePointsMap = createFeaturePointMap(
                ProcrustesUtils.sortListByFeaturePointType(modifiableFeaturePointList));
        this.featurePointTypeCorrespondence = createFeaturePointTypeCorrespondence(face.getFeaturePoints());
    }


    /**
     * Sets feature points map and also sets feature point on human face for
     * visualisation.
     *
     * @param featurePointsMap
     */
    public void setFeaturePointsMap(HashMap<Integer, FeaturePoint> featurePointsMap) {
        this.featurePointsMap = featurePointsMap;
//        readjustFeaturePoints(featurePointsMap);
//        this.humanFace.setFeaturePoints(getFeaturePointValues());
    }

    public List<FeaturePoint> getFeaturePointValues() {
        return new ArrayList<>(this.getFeaturePointsMap().values());
    }

    public HashMap<Integer, FeaturePoint> getFeaturePointsMap() {
        return featurePointsMap;
    }

    public HashMap<Integer, Integer> getFeaturePointTypeCorrespondence() {
        return featurePointTypeCorrespondence;
    }

    /**
     * Creates a subset of selected feature points that can be used for
     * procrustes analysis. Each of the feature point types contained in the
     * subset is contained in both faces.
     *
     * @param featurePoints
     * @param viablePoints
     * @return
     */
    private List<FeaturePoint> getSelectionOfFeaturePoints(List<FeaturePoint> featurePoints, List<Integer> viablePoints) {
        ArrayList<FeaturePoint> selection = new ArrayList<>();
        featurePoints.forEach(fp -> {
            if (viablePoints.contains(fp.getFeaturePointType().getType())) {
                selection.add(fp);
            }
        });
        return selection;
    }

    /**
     * Creates corresponding key value pair of matrix row index and a feature
     * point type value. It is used so we can get back type of feature point
     * after working with matrices where the type is not stored, but instead are
     * used matrix indexes.
     *
     * @param fpList
     * @return
     */
    private HashMap<Integer, Integer> createFeaturePointTypeCorrespondence(List<FeaturePoint> fpList) {
        HashMap<Integer, Integer> map = new HashMap<>();
        for (int i = 0; i < fpList.size(); i++) {
            map.put(i, fpList.get(i).getFeaturePointType().getType());
        }
        return map;
    }

    /**
     * Creates feature point hash map for internal use in Procrustes Analysis.
     *
     * Map key is feature point type (Integer). Value is feature point itself.
     *
     * @param featurePointList
     * @return
     */
    private HashMap<Integer, FeaturePoint> createFeaturePointMap(List<FeaturePoint> featurePointList) {
        HashMap<Integer, FeaturePoint> map = new HashMap<>();
        for (FeaturePoint fp : featurePointList) {
            map.put(fp.getFeaturePointType().getType(), fp);
        }
        return map;
    }

//    private void readjustFeaturePoints(HashMap<Integer, FeaturePoint> featurePointsMap) {
////        if (featurePointsMap.size() != this.humanFace.getFeaturePoints().size()) {
////            readjustSelection(featurePointsMap);
////        } else {
//        this.humanFace.getFeaturePoints().forEach(fp -> {
//            FeaturePoint movedFp = featurePointsMap.get(fp.getFeaturePointType().getType());
//            if (movedFp != null) {
//                if (fp.getFeaturePointType().getType()
//                        != movedFp.getFeaturePointType().getType()) {
//                    throw new RuntimeException("Types do not correspond");
//                }
//                fp.getPosition().x = movedFp.getX();
//                fp.getPosition().y = movedFp.getY();
//                fp.getPosition().z = movedFp.getZ();
//            }
//        });
////    }
//    }

}
